--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   Transform.hs
--
-- Author:
--   Klaus Aehlig (2010)
--
-- Description:
--    Compute the let- and sharing normal form of a term


module Transform where


import RAMLTypes
import Util
import List
import Control.Monad.State.Strict




--------------------------------------------------------------------------------
-- Unification of function declarations:
-- Every function should have exactly one argument!
-- Expressions should be in Share-Let-Normal-Form

normalizeFunDecs typedFunList = [ map normalizeDec funSCC | funSCC <- typedFunList ]
   where
     normalizeDec (f,(context,e_f)) = 
         let (newVar:newVars) = [ ('^':f)++(show i) | i <- [1..] ]
             ef_normal = shareLetNormal newVars e_f
         in
         case context of
           [(x,t)] -> (f,(context, ef_normal))
           _       -> let (xs,ts) = unzip context
                          t = TTuple ts
                          e_f' = TEMatchT Free (TEVar newVar t) xs 
                                               (ef_normal) (typeOf e_f)
                      in (f,([(newVar,t)],e_f') )





--------------------------------------------------------------------------------
-- Substitutions

-- Substitute expressions for variables in expressions without explicit sharing.

substExp :: TExp -> VarId -> TExp -> TExp
          -- e   ->  y    -> e'     -> e[y:=e']
substExp ex y eSub = sub ex
   where
     sub ex = 
         case ex of
           TEVar x t   -> if x==y then eSub else (TEVar x t)
           TEApp f e t -> TEApp f (sub e) t
           TEUnaryOp op e t -> TEUnaryOp op (sub e) t
           TEBinOp op e1 e2 t -> TEBinOp op (sub e1) (sub e2) t
           TESucc e t -> TESucc (sub e) t
           TEToInt e t -> TEToInt (sub e) t
           TEMatchN e e1 x e2 t -> TEMatchN (sub e) (sub e1) x (subBound [x] e2) t
           TELet kind x e1 e2 t -> TELet kind x (sub e1) (subBound [x] e2) t
           TECond e e1 e2 t -> TECond (sub e) (sub e1) (sub e2) t
           TECons kind e_h e_t t -> TECons kind (sub e_h) (sub e_t) t
           TETuple es t -> TETuple (map sub es) t
           TEMatchL mt e e1 x1 x2 e2 t -> TEMatchL mt (sub e) (sub e1) x1 x2 (subBound [x1,x2] e2) t
           TEMatchT mt e1 xs e2 t -> TEMatchT mt (sub e1) xs (subBound xs e2) t
           TENodeB kind e e1 e2 t -> TENodeB kind (sub e) (sub e1) (sub e2) t
           TEMatchB mt e e1 x1 x2 x3 e2 t -> TEMatchB mt (sub e) (sub e1) x1 x2 x3 (subBound [x1,x2,x3] e2) t
           TEShare x1 x2 x e t -> error "Trying to substitute an expression for a variable in a share expression."
           e -> e     -- nil, leaf, unit, true, false, ...

     subBound xs e = if y `elem` xs then e else sub e


-- Substitute variables in arbritary expressions. 
subst :: TExp -> VarId -> VarId -> TExp
      -- e   ->  y    -> z     -> e[y:=z]

subst ex y z = sub ex
   where
     sub ex = 
         case ex of
           TEVar x t   -> TEVar (if x==y then z else x) t
           TEApp f e t -> TEApp f (sub e) t
           TEUnaryOp op e t -> TEUnaryOp op (sub e) t
           TEBinOp op e1 e2 t -> TEBinOp op (sub e1) (sub e2) t
           TESucc e t -> TESucc (sub e) t
           TEToInt e t -> TEToInt (sub e) t
           TEMatchN e e1 x e2 t -> TEMatchN (sub e) (sub e1) x (subBound [x] e2) t
           TELet kind x e1 e2 t -> TELet kind x (sub e1) (subBound [x] e2) t
           TECond e e1 e2 t -> TECond (sub e) (sub e1) (sub e2) t
           TECons kind e_h e_t t -> TECons kind (sub e_h) (sub e_t) t
           TETuple es t -> TETuple (map sub es) t
           TEMatchL mt e e1 x1 x2 e2 t -> TEMatchL mt (sub e) (sub e1) x1 x2 (subBound [x1,x2] e2) t
           TEMatchT mt e1 xs e2 t -> TEMatchT mt (sub e1) xs (subBound xs e2) t
           TENodeB kind e e1 e2 t -> TENodeB kind (sub e) (sub e1) (sub e2) t
           TEMatchB mt e e1 x1 x2 x3 e2 t -> TEMatchB mt (sub e) (sub e1) x1 x2 x3 (subBound [x1,x2,x3] e2) t
           TEShare x1 x2 x e t -> TEShare x1 x2 (if x==y then z else x) (subBound [x1,x2] e) t
           e -> e     -- nil, leaf, unit, true, false, ...

     subBound xs e = if y `elem` xs then e else sub e

getVar :: State [String] String

getVar = do { nv:nvs <- get
            ; put nvs
            ; return nv
            }

getVars n = sequence [ getVar | i <- [1..n] ]


--------------------------------------------------------------------------------
-- Sharing, in CPS
--------------------------------------------------------------------------------


shareVar x e1 es cont = 
    do { nv1 <- getVar
       ; nv2 <- getVar
       ; result <- cont (subst e1 x nv1) (map (\e -> subst e x nv2) es)
       ; return $ TEShare nv1 nv2 x result (typeOf result)
       }

shareVars vars e1 es cont = case vars of
                              []     -> cont e1 es
                              (x:xs) -> shareVar x e1 es (\e1' -> \es' -> shareVars xs e1' es' cont)

shareVarsTuple es cont = 
    case es of
      [] -> cont []
      (e:es) -> shareVars (freeVars e `intersect` (foldl union [] (map freeVars es)) ) e es
                (\e' es' -> shareVarsTuple es' (\es'' -> cont $ e':es'') )



shareFreeVars e1 e2 cont = 
    shareVars ((freeVars e1) `intersect` (freeVars e2)) e1 [e2] (\e1' [e2'] -> cont e1' e2')



--------------------------------------------------------------------------------

freeLets lets exp = case lets of
                      [] -> return exp
                      (x,e):lets' -> do { exp' <- freeLets lets' exp
                                        ; return $ TELet Free x e exp' (typeOf exp')
                                        }
                   


--------------------------------------------------------------------------------


shareLetNormal :: [VarId] -> TExp -> TExp

shareLetNormal varNames e = evalState (norm e) varNames


norm exp = 
    if typeOf exp == TNatConst then fail $ "Error: Type TNatConst wasn't eliminated!\n In Expression \n" ++ (show exp)  -- XXX sanity check
    else
    case exp of
       TEApp f e t ->       -- do we need a fresh variable here ???
          do { e' <- norm e
             ; return $ TEApp f e' t
             }

       TEUnaryOp op e t ->
          do { e' <- norm e
             ; return $ TEUnaryOp op e' t
             }

       TEBinOp b e1 e2 t ->
          do { e1' <- norm e1
             ; e2' <- norm e2
             ; [nv1,nv2] <- getVars 2
             ; shareFreeVars e1' e2' (\ e1' e2' -> 
               freeLets [(nv1,e1'),(nv2,e2')]
               (TEBinOp b (TEVar nv1 (typeOf e1')) (TEVar nv2 (typeOf e2')) t) )
             }

       TESucc e t -> 
          do { e' <- norm e
             ; nv <- getVar
             ; freeLets [(nv,e')] $ TESucc (TEVar nv (typeOf e')) t
             }

       TEToInt e t -> 
          do { e' <- norm e
             ; nv <- getVar
             ; freeLets [(nv,e')] $ TEToInt (TEVar nv (typeOf e')) t
             }

       TEMatchN e e1 x e2 t ->
           do { e' <- norm e
              ; e1' <- norm e1
              ; e2' <- norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` 
                                 ((freeVars e1') `union` ((freeVars e2') \\ [x]))
              ; shareVars commonVars e' [e1',e2'] (\ e' [e1',e2'] ->
                freeLets [(nv,e')] $ TEMatchN (TEVar nv (typeOf e')) e1' x e2' t)
              } 

       TELet kind x e1 e2 t ->
           do { e1' <- norm e1
              ; e2' <- norm e2
              ; let commonVars = ((freeVars e1') `intersect` ((freeVars e2) \\ [x]))
              ; shareVars commonVars e1' [e2'] (\ e1' [e2'] ->
                return $ TELet kind x e1' e2' t )
              }

       TEShare x1 x2 x e t ->
           do { e' <- norm e 
              ; return $ TEShare x1 x2 x e' t
              }

       TECond e e1 e2 t ->
           do { e' <- norm e
              ; e1' <- norm e1
              ; e2' <- norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` ((freeVars e1') `union` (freeVars e2'))
              ; shareVars commonVars e' [e1',e2'] (\e' [e1',e2'] ->
                freeLets [(nv,e')] $ TECond (TEVar nv (typeOf e')) e1' e2' t )
              }

       TECons Free (TEVar x t) (TEVar _ _) _ -> freeLets [(x,TEVar x t)] exp  -- The freelet is to count a variable
                                                                              -- in the analysis.

       TECons kind e1 e2 t ->
           do { e1' <- norm e1
              ; e2' <- norm e2
              ; [nv1,nv2] <- getVars 2
              ; shareFreeVars e1' e2' (\ e1' e2' -> 
                  freeLets [(nv1,e1'),(nv2,e2')] $
                     TECons kind (TEVar nv1 (typeOf e1')) (TEVar nv2 (typeOf e2')) t )
              }

       TEMatchL mt e e1 x1 x2 e2 t ->
           do { e' <- norm e
              ; e1' <- case e of 
                         TEVar y listT -> norm $ substExp e1 y (TENil Free listT)
                         _ -> norm e1
              ; e2' <- case e of
                         TEVar y listT@(TList elemT) -> 
                             if y `elem` [x1,x2] then norm e2 else  -- Watch out with shadowing of y
                             norm $ substExp e2 y (TECons Free (TEVar x1 elemT) (TEVar x2 listT) listT)
                         _ -> norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` 
                                 ((freeVars e1') `union` ((freeVars e2') \\ [x1,x2]))
              ; shareVars commonVars e' [e1',e2'] (\ e' [e1',e2'] ->
                freeLets [(nv,e')] $ TEMatchL mt (TEVar nv (typeOf e')) e1' x1 x2 e2' t)
              } 

       TETuple es t ->
           do { es' <- sequence [ norm e | e <- es]
              ; nvs <- getVars (length es)
              ; shareVarsTuple es' (\ es' ->
                let varsEs = zip nvs es' in
                freeLets varsEs  $ TETuple (map (\(nv,e) -> TEVar nv (typeOf e)) varsEs ) t )
              }

       TEMatchT mt e xs e1 t ->
           do { e' <- norm e
              ; e1'  <- norm e1
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` ((freeVars e1') \\ xs)
              ; shareVars commonVars e' [e1'] (\ e' [e1'] ->
                freeLets [(nv,e')] $ TEMatchT mt (TEVar nv (typeOf e')) xs e1' t )
              }
 
       TENodeB Free (TEVar x t) (TEVar _ _) (TEVar _ _) _ -> freeLets [(x,TEVar x t)] exp  -- just to count KVar

       TENodeB kind e1 e2 e3 t ->
           do { es' <- sequence [ norm e | e <- [e1,e2,e3] ]
              ; nvs <- getVars 3
              ; shareVarsTuple es' (\ es' ->
                let varsEs = zip nvs es'
                    [v1,v2,v3] = map (\(nv,e) -> TEVar nv (typeOf e)) varsEs
                in
                freeLets varsEs  $ TENodeB kind v1 v2 v3 t )
              }

       TEMatchB mt e e1 x1 x2 x3 e2 t ->
           do { e' <- norm e
              ; e1' <- case e of
                         TEVar y treeT -> 
                             norm $ substExp e1 y (TELeaf Free treeT)
                         _ -> norm e1
              ; e2' <- case e of
                         TEVar y treeT@(TList elemT) -> 
                             if y `elem` [x1,x2,x3] then norm e2 else  -- Watch out with shadowing of y
                             norm $ substExp e2 y (TENodeB Free (TEVar x1 elemT) (TEVar x2 elemT) (TEVar x3 treeT) treeT)
                         _ -> norm e2
              ; nv <- getVar
              ; let commonVars = (freeVars e') `intersect` 
                                 ((freeVars e1') `union` ((freeVars e2') \\ [x1,x2,x3]))
              ; shareVars commonVars e' [e1',e2'] (\ e' [e1',e2'] ->
                freeLets [(nv,e')] $ TEMatchB mt (TEVar nv (typeOf e')) e1' x1 x2 x3 e2' t)
              } 

--  TEUnit TETrue TEFalse TENat TEVar TENil TELeaf TETick
       _ -> return exp

